/**************************************************************************
 *
 * This program is part of Zenoss Core, an open source monitoring platform.
 * Copyright (C) 2007, Zenoss Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * For complete information please visit: http://www.zenoss.com/oss/
 *
 **************************************************************************/
#include <sys/types.h>

#include <getopt.h>
#include <ifaddrs.h>
#include <netdb.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <sys/socket.h>

#include <netinet/in.h>


static struct option options[] = {
    {"ping",   no_argument, 0, 0},
    {"listen", no_argument, 0, 0},
    {"proto", required_argument, 0, 0},
    {"port", required_argument, 0, 0},
    { 0, 0, 0, 0 }
};

static void error(const char * msg) {
    printf("ZenSocket Error: %s\n", msg);
    exit(EXIT_FAILURE);
}

static void usage(const char *name) {
  printf("\n");
  printf("Usage: %s [options] -- command to run\n", name);
  printf("\n");
  printf("Send Options:\n");
  printf("  --ping                       send ICMP packets\n");
  printf("\n");
  printf("Listen Options:\n");
  printf("  --listen                     listen for incoming packets\n");
  printf("  --port:[interface:]number    the numerical port to listen on\n");
  printf("  --proto:[udp|tcp]            protocol to listen for\n");
  printf("\n");
  exit(EXIT_FAILURE);
}

static in_addr_t getAddress(const char *interface) {
  int a, b, c, d;

  if (!interface || !*interface) {
    return INADDR_ANY;
  }

  if (sscanf(interface, "%d.%d.%d.%d", &a, &b, &c, &d) != 4) {
    struct addrinfo *res = 0, *p;

    if (getaddrinfo(interface, NULL, NULL, &res) == 0) {
	    in_addr_t result;
	    for (p = res; p; p = p->ai_next) {
        if (p->ai_family == AF_INET) {
          struct sockaddr_in * in = (struct sockaddr_in *)p->ai_addr;
          result = in->sin_addr.s_addr;
          break;
        }
	    }

	    freeaddrinfo(res);
	    if (p) {
        return result;
      }
    }

    fprintf(stderr, "Warning: unable to decode address %s\n", interface);
    return INADDR_ANY;
  }

  return htonl( ((a & 0xff) << 24) | ((b & 0xff) << 16) | 
                ((c & 0xff) << 8)  |  (d & 0xff) );
}

int main(int argc, char **argv) {
  char **args = 0;
  const char * replace = "$privilegedSocket";
  int i, c;
  int ping = 0;
  int listen = 0;
  char * proto;
  const char * udp = "udp";
  const char * tcp = "tcp";
  const char * interface = 0;
  unsigned short port = 0;
  int sock = -1;
  char filenoString[10] = "";
  
  if (geteuid() != 0) {
    error("zensocket needs to be run as root or setuid");
  }

  if (argc < 2) {
    usage("zensocket");
  }

  
  while (1) {
    int option_index = 0;
    int c = getopt_long(argc, argv, "", options, &option_index);
    if (c == -1)
	    break;
    switch (c) {
    case 0:
      /* ping option */
	    if (option_index == 0)
        ping = 1;

      /* listen option */
	    if (option_index == 1) {
        listen = 1;
      }

      /* protocol option */
      if (option_index == 2) {
        proto = optarg;
      }

      /* port option */
      if (option_index == 3) {
        char * portString = optarg;
        char * colon = strchr(optarg, ':');
        if (colon) {
          *colon = '\0';
          interface = optarg;
          portString = colon + 1;
        }
   
        port = atol(portString);
        if (port == 0) {
          error("Unable to use port number");
        }
	    }

	    break;

    default:
	    usage(argv[0]);
    }
  }

  if (!ping && !listen) {
    usage(argv[0]);
  }

  if (ping) {
    sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if (sock < 0)
	    error("Unable to create ping socket");
  } 

  if (listen) {
    int reuse_addr = 1;
    struct sockaddr_in addr;
    int proto_sock = SOCK_DGRAM;
    
    /* if the proto option was udp, use SOCK_DGRAM */
    if (! strcmp(proto, udp)) {
      proto_sock = SOCK_DGRAM;
    }

    /* if the proto option was tcp, use SOCK_STREAM */
    if (! strcmp(proto, tcp)) {
      proto_sock = SOCK_STREAM;
    }

    sock = socket(AF_INET, proto_sock, 0);
    if (sock < 0) {
	    error("Unable to create listen socket");
    }

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = getAddress(interface);
    addr.sin_port = htons(port);
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, 
                   &reuse_addr, sizeof(reuse_addr)) < 0) {
	    error("Unable to setsockopt");
    }


    if (bind(sock, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
      char buffer[50];
      sprintf(buffer, "Unable to bind to listen port (error code %d)", errno);
	    error(buffer);
    }
  }

  /* No need to be root any more */
  {
    struct passwd * p = NULL;
    while (1) {
	    p = getpwent();

	    if (!p || strcmp(p->pw_name, "zenoss") == 0) {
        break;
      }
    }

    if (p) {
	    setgid(p->pw_gid);
	    setuid(p->pw_uid);
	    setenv("HOME", p->pw_dir, 1);
    } else {
	    if (getuid() != 0) {
        setuid(getuid());
	    } else {
        error("Unable to find a user to become");
	    }
    }
    endpwent();
  }

  if (optind >= argc) {
    usage(argv[0]);
  }
  
  sprintf(filenoString, "%d", sock);
  if (strlen(filenoString) > strlen(replace)) {
    error("What are you up to?");
  }
  args = (char**)malloc(sizeof(char*)*(argc - optind + 1));

  if (!args) {
    error("Out of memory");
  }

  for (i = optind, c = 0; i < argc; i++, c++) {
    char * p = strstr(argv[i], replace);
    if (p) {
	    strcpy(p, filenoString);
	    strcpy(p + strlen(filenoString), p + strlen(replace));
    }

    args[c] = argv[i];
  }

  args[c] = 0;
  {
      char libdir[1000];
      snprintf(libdir, sizeof(libdir)-1, "LD_LIBRARY_PATH=%s/lib", getenv("ZENHOME"));
      if (putenv(libdir) != 0) {
	  perror("Unable to set LD_LIBRARY_PATH");
      }
      if (execv(args[0], args) < 0) {
	perror("exec fails");
      }
  }

  free(args);
  exit(EXIT_SUCCESS);

  return 0;
}
